"""
Radiation and equilibrium emittances
"""
from math import sin, cos, tan, sqrt, sinh, cosh, pi
import numpy
from scipy.linalg import inv, det, solve_sylvester
from scipy.constants import c as clight
from at.lattice import Lattice, check_radiation, uint32_refpts
from at.lattice import Element, elements
from at.tracking import lattice_pass
from at.physics import find_orbit6, find_m66, find_elem_m66, get_tunes_damp
from at.physics import Cgamma, linopt, find_mpole_raddiff_matrix, e_mass

__all__ = ['ohmi_envelope', 'get_radiation_integrals', 'quantdiffmat',
           'gen_quantdiff_elem']

_NSTEP = 60  # nb slices in a wiggler period

_submat = [slice(0, 2), slice(2, 4), slice(6, 3, -1)]

# dtype for structured array containing optical parameters
ENVELOPE_DTYPE = [('r66', numpy.float64, (6, 6)),
                  ('r44', numpy.float64, (4, 4)),
                  ('m66', numpy.float64, (6, 6)),
                  ('orbit6', numpy.float64, (6,)),
                  ('emitXY', numpy.float64, (2,)),
                  ('emitXYZ', numpy.float64, (3,))]


def _cumulb(it):
    """accumulate diffusion matrices"""
    cumul = numpy.zeros((6, 6))
    yield cumul
    for el, orbin, b in it:
        m = find_elem_m66(el, orbin)
        cumul = m.dot(cumul).dot(m.T) + b
        yield cumul


def _dmatr(ring, orbit=None, keep_lattice=False):
    """
    compute the cumulative diffusion and orbit
    matrices over the ring
    """
    nelems = len(ring)
    energy = ring.energy
    allrefs = uint32_refpts(range(nelems + 1), nelems)

    if orbit is None:
        orbit, _ = find_orbit6(ring, keep_lattice=keep_lattice)
        keep_lattice = True

    orbs = numpy.squeeze(
        lattice_pass(ring, orbit.copy(order='K'), refpts=allrefs,
                     keep_lattice=keep_lattice), axis=(1, 3)).T
    b0 = numpy.zeros((6, 6))
    bb = [find_mpole_raddiff_matrix(elem, elemorb, energy)
          if elem.PassMethod.endswith('RadPass') else b0
          for elem, elemorb in zip(ring, orbs)]
    bbcum = numpy.stack(list(_cumulb(zip(ring, orbs, bb))), axis=0)
    return bbcum, orbs


def _lmat(dmat):
    """
    lmat is Cholesky decomp of dmat unless diffusion is 0 in
    vertical.  Then do chol on 4x4 hor-long matrix and put 0's
    in vertical
    """
    lmat = numpy.zeros((6, 6))
    try:
        lmat = numpy.linalg.cholesky(dmat)
    except numpy.linalg.LinAlgError:
        nz = numpy.where(dmat != 0)
        cmat = numpy.reshape(dmat[nz], (4, 4))
        cmat = numpy.linalg.cholesky(cmat)
        lmat[nz] = numpy.reshape(cmat, (16,))
    return lmat


@check_radiation(True)
def ohmi_envelope(ring, refpts=None, orbit=None, keep_lattice=False):
    """
    Calculate the equilibrium beam envelope in a
    circular accelerator using Ohmi's beam envelope formalism [1]

    emit0, beamdata, emit = ohmi_envelope(ring[, refpts])

    PARAMETERS
        ring            Lattice object.
        refpts=None     elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are True.

    KEYWORDS
        orbit=None          Avoids looking for the closed orbit if it is
                            already known ((6,) array)
        keep_lattice=False  Assume no lattice change since the previous
                            tracking

    OUTPUT
        emit0               emittance data at the start/end of the ring
        beamdata            beam parameters at the start of the ring
        emit                emittance data at the points refered to by refpts,
                            if refpts is None an empty structure is returned.

        emit is a record array with fields:
        r66                 (6, 6) equilibrium envelope matrix R
        r44                 (4, 4) betatron emittance matrix (dpp = 0)
        m66                 (6, 6) transfer matrix from the start of the ring
        orbit6              (6,) closed orbit
        emitXY              (2,) betatron emittance projected on xxp and yyp
        emitXYZ             (3,) 6x6 emittance projected on xxp, yyp, ldp

        beamdata is a record array with fields:
        tunes               tunes of the 3 normal modes
        damping_rates       damping rates of the 3 normal modes
        mode_matrices       R-matrices of the 3 normal modes
        mode_emittances     equilibrium emittances of the 3 normal modes

        Field values can be obtained with either
        emit['r66']    or
        emit.r66

    REFERENCES
        [1] K.Ohmi et al. Phys.Rev.E. Vol.49. (1994)
    """

    def process(r66):
        # projections on xx', zz', ldp
        emit3sq = numpy.array([det(r66[s, s]) for s in _submat])
        # Prevent from unrealistic negative values of the determinant
        emit3 = numpy.sqrt(numpy.maximum(emit3sq, 0.0))
        # Emittance cut for dpp=0
        if emit3[0] < 1.E-13:  # No equilibrium emittance
            r44 = numpy.nan * numpy.ones((4, 4))
        elif emit3[1] < 1.E-13:  # Uncoupled machine
            minv = inv(r66[[0, 1, 4, 5], :][:, [0, 1, 4, 5]])
            r44 = numpy.zeros((4, 4))
            r44[:2, :2] = inv(minv[:2, :2])
        else:  # Coupled machine
            minv = inv(r66)
            r44 = inv(minv[:4, :4])
        # betatron emittances (dpp=0)
        emit2sq = numpy.array(
            [det(r44[s, s], check_finite=False) for s in _submat[:2]])
        # Prevent from unrealistic negative values of the determinant
        emit2 = numpy.sqrt(numpy.maximum(emit2sq, 0.0))
        return r44, emit2, emit3

    def propag(m, cumb, orbit6):
        """Propagate the beam matrix to refpts"""
        sigmatrix = m.dot(rr).dot(m.T) + cumb
        m44, emit2, emit3 = process(sigmatrix)
        return sigmatrix, m44, m, orbit6, emit2, emit3

    nelems = len(ring)
    uint32refs = uint32_refpts(refpts, nelems)
    bbcum, orbs = _dmatr(ring, orbit=orbit, keep_lattice=keep_lattice)
    mring, ms = find_m66(ring, uint32refs, orbit=orbs[0], keep_lattice=True)
    # ------------------------------------------------------------------------
    # Equation for the moment matrix R is
    #         R = MRING*R*MRING' + BCUM;
    # We rewrite it in the form of Lyapunov-Sylvester equation to use scipy's
    # solve_sylvester function
    #            A*R + R*B = Q
    # where
    #               A =  inv(MRING)
    #               B = -MRING'
    #               Q = inv(MRING)*BCUM
    # ------------------------------------------------------------------------
    aa = inv(mring)
    bb = -mring.T
    qq = numpy.dot(aa, bbcum[-1])
    rr = solve_sylvester(aa, bb, qq)
    rr = 0.5 * (rr + rr.T)
    rr4, emitxy, emitxyz = process(rr)
    r66data = get_tunes_damp(mring, rr)

    data0 = numpy.rec.fromarrays(
        (rr, rr4, mring, orbs[0], emitxy, emitxyz),
        dtype=ENVELOPE_DTYPE)
    if uint32refs.shape == (0,):
        data = numpy.recarray((0,), dtype=ENVELOPE_DTYPE)
    else:
        data = numpy.rec.fromrecords(
            list(map(propag, ms, bbcum[uint32refs], orbs[uint32refs, :])),
            dtype=ENVELOPE_DTYPE)

    return data0, r66data, data


@check_radiation(False)
def get_radiation_integrals(ring, dp=0.0, twiss=None):
    """
    Compute the 5 radiation integrals for uncoupled lattices. No RF cavity or
    radiating element is allowed.

    PARAMETERS
        ring            lattice description.
        dp=0.0          momentum deviation

    KEYWORDS
        twiss=None      linear optics at all points (from linopt). If None,
                        it will be computed.

    OUTPUT
        i1, i2, i3, i4, i5
    """

    def dipole_radiation(elem, vini, vend):
        """Analytically compute the radiation integrals in dipoles"""
        beta0 = vini.beta[0]
        alpha0 = vini.alpha[0]
        eta0 = vini.dispersion[0]
        etap0 = vini.dispersion[1]

        ll = elem.Length
        theta = elem.BendingAngle
        rho = ll / theta
        rho2 = rho * rho
        k2 = elem.K + 1.0 / rho2
        eps1 = tan(elem.EntranceAngle) / rho
        eps2 = tan(elem.ExitAngle) / rho

        eta3 = vend.dispersion[0]
        alpha1 = alpha0 - beta0 * eps1
        gamma1 = (1.0 + alpha1 * alpha1) / beta0
        etap1 = etap0 + eta0 * eps1
        etap2 = vend.dispersion[1] - eta3 * eps2

        h0 = gamma1*eta0*eta0 + 2.0*alpha1*eta0*etap1 + beta0*etap1*etap1

        if k2 != 0.0:
            if k2 > 0.0:  # Focusing
                kl = ll * sqrt(k2)
                ss = sin(kl) / kl
                cc = cos(kl)
            else:  # Defocusing
                kl = ll * sqrt(-k2)
                ss = sinh(kl) / kl
                cc = cosh(kl)
            eta_ave = (theta - (etap2 - etap1)) / k2 / ll
            bb = 2.0 * (alpha1 * eta0 + beta0 * etap1) * rho
            aa = -2.0 * (alpha1 * etap1 + gamma1 * eta0) * rho
            h_ave = h0 + (aa * (1.0 - ss) + bb * (1.0 - cc) / ll
                          + gamma1 * (3.0 - 4.0 * ss + ss * cc) / 2.0 / k2
                          - alpha1 * (1.0 - cc) ** 2 / k2 / ll
                          + beta0 * (1.0 - ss * cc) / 2.0
                          ) / k2 / rho2
        else:
            eta_ave = 0.5 * (eta0 + eta3) - ll * ll / 12.0 / rho
            hp0 = 2.0 * (alpha1 * eta0 + beta0 * etap1) / rho
            h2p0 = 2.0 * (-alpha1 * etap1 + beta0 / rho - gamma1 * eta0) / rho
            h_ave = h0 + hp0 * ll / 2.0 + h2p0 * ll * ll / 6.0 \
                    - alpha1 * ll ** 3 / 4.0 / rho2 \
                    + gamma1 * ll ** 4 / 20.0 / rho2

        di1 = eta_ave * ll / rho
        di2 = ll / rho2
        di3 = ll / abs(rho) / rho2
        di4 = eta_ave * ll * (2.0 * elem.K + 1.0 / rho2) / rho \
              - (eta0 * eps1 + eta3 * eps2) / rho
        di5 = h_ave * ll / abs(rho) / rho2
        return numpy.array([di1, di2, di3, di4, di5])

    def wiggler_radiation(elem, dini):
        """Compute the radiation integrals in wigglers with the following
        approximations:

        - The wiggler is aligned with the closed orbit
        - The self-induced dispersion is neglected in I4 and I5, but is is used
          as a lower limit for the I5 contribution
        - I1, I2 are integrated analytically
        - I3 is integrated analytically for a single harmonic, numerically
          otherwise
        """

        def b_on_axis(wig, s):
            """On-axis wiggler field"""

            def harm(coef, h, phi):
                return -Bmax * coef * numpy.cos(h*kws + phi)

            kw = 2 * pi / wig.Lw
            Bmax = wig.Bmax
            kws = kw * s
            zz = [numpy.zeros(kws.shape)]
            vh = zz + [harm(pb[1], pb[4], pb[5]) for pb in wig.By.T]
            vv = zz + [-harm(pb[1], pb[4], pb[5]) for pb in wig.Bx.T]
            bys = numpy.sum(numpy.stack(vh), axis=0)
            bxs = numpy.sum(numpy.stack(vv), axis=0)
            return bxs, bys

        le = elem.Length
        alphax0 = dini.alpha[0]
        betax0 = dini.beta[0]
        gammax0 = (alphax0 * alphax0 + 1) / betax0
        eta0 = dini.dispersion[0]
        etap0 = dini.dispersion[1]
        H0 = gammax0*eta0*eta0 + 2*alphax0*eta0*etap0 + betax0*etap0*etap0
        avebetax = betax0 + alphax0*le + gammax0*le*le/3

        kw = 2 * pi / elem.Lw
        rhoinv = elem.Bmax / Brho
        coefh = elem.By[1, :] * rhoinv
        coefv = elem.Bx[1, :] * rhoinv
        coef2 = numpy.concatenate((coefh, coefv))
        if len(coef2) == 1:
            di3 = le * coef2[0] ** 3 * 4 / 3 / pi
        else:
            bx, bz = b_on_axis(elem, numpy.linspace(0, elem.Lw, _NSTEP + 1))
            rinv = numpy.sqrt(bx*bx + bz*bz) / Brho
            di3 = numpy.trapz(rinv ** 3) * le / _NSTEP
        di2 = le * (numpy.sum(coefh * coefh) + numpy.sum(coefv * coefv)) / 2
        di1 = -di2 / kw / kw
        di4 = 0
        if len(coefh) > 0:
            d5lim = 4 * avebetax * le * coefh[0] ** 5 / 15 / pi / kw / kw
        else:
            d5lim = 0
        di5 = max(H0 * di3, d5lim)
        return numpy.array([di1, di2, di3, di4, di5])

    Brho = sqrt(ring.energy**2 - e_mass**2) / clight
    integrals = numpy.zeros((5,))

    if twiss is None:
        _, _, _, twiss = linopt(ring, dp, range(len(ring) + 1),
                                get_chrom=True, coupled=False)
    elif len(twiss) != len(ring) + 1:
        raise ValueError('length of Twiss data should be {0}'
                         .format(len(ring) + 1))
    for (el, vini, vend) in zip(ring, twiss[:-1], twiss[1:]):
        if isinstance(el, elements.Dipole) and el.BendingAngle != 0.0:
            integrals += dipole_radiation(el, vini, vend)
        elif isinstance(el, elements.Wiggler) and el.PassMethod != 'DriftPass':
            integrals += wiggler_radiation(el, vini)
    return tuple(integrals)


def get_energy_loss(ring):
    """Energy loss per turn [eV]

    Losses = Cgamma / 2pi * EGeV^4 * i2
    """

    def wiggler_i2(el):
        rhoinv = el.Bmax / Brho
        coefh = el.By[1, :]
        coefv = el.Bx[1, :]
        return el.Length * (numpy.sum(coefh * coefh) + numpy.sum(
            coefv*coefv)) * rhoinv ** 2 / 2

    def dipole_i2(el):
        return el.BendingAngle ** 2 / el.Length

    Brho = sqrt(ring.energy**2 - e_mass**2) / clight
    i2 = 0.0
    for el in ring:
        if isinstance(el, elements.Dipole):
            i2 += dipole_i2(el)
        elif isinstance(el, elements.Wiggler) and el.PassMethod != 'DriftPass':
            i2 += wiggler_i2(el)
    i2 = ring.periodicity * i2
    e_loss = Cgamma / 2.0 / pi * ring.energy ** 4 * i2
    return e_loss


@check_radiation(True)
def quantdiffmat(ring, orbit=None):
    """
    This function computes the diffusion matrix of the whole ring

    PARAMETERS
        ring            lattice description.
        orbit=None      initial orbit

    OUTPUT
        diffusion matrix (6,6)
    """
    bbcum, _ = _dmatr(ring, orbit=orbit)
    diffmat = [(bbc + bbc.T) / 2 for bbc in bbcum]
    return numpy.round(diffmat[-1], 24)


@check_radiation(True)
def gen_quantdiff_elem(ring, orbit=None):
    """
    Generates a quantum diffusion element
    """
    dmat = quantdiffmat(ring, orbit=orbit)
    lmat = numpy.asfortranarray(_lmat(dmat))
    diff_elem = Element('Diffusion',
                        Lmatp=lmat,
                        PassMethod='QuantDiffPass')
    return diff_elem


Lattice.ohmi_envelope = ohmi_envelope
Lattice.get_radiation_integrals = get_radiation_integrals
Lattice.energy_loss = property(get_energy_loss)



"""
Function to compute quantities related to non-linear optics
"""
import numpy
from math import sqrt, atan2
from scipy.special import factorial
from at.lattice import Lattice, check_radiation, uint32_refpts, get_s_pos, \
    bool_refpts
from at.tracking import lattice_pass
from at.physics import HarmonicAnalysis, get_tune, linopt, find_orbit4, \
    get_tunes_harmonic, get_chrom
from at.lattice import Element

__all__ = ['detuning', 'chromaticity', 'gen_detuning_elem']


def tunes_vs_amp(ring, amp=None, dim=0,
                 dp=0, window=1, nturns=512):
    """
    Generates a range of tunes for varying x, y, or z amplitudes
    """

    def _gen_part(ring, amp, dim, orbit, ld, nturns):
        invx = numpy.array([[1/numpy.sqrt(ld['beta'][0]), 0],
                            [ld['alpha'][0]/numpy.sqrt(ld['beta'][0]),
                            numpy.sqrt(ld['beta'][0])]])

        invy = numpy.array([[1/numpy.sqrt(ld['beta'][1]), 0],
                            [ld['alpha'][1]/numpy.sqrt(ld['beta'][1]),
                            numpy.sqrt(ld['beta'][1])]])
        part = numpy.array([orbit, ] * len(amp)).T + 1.0e-6
        part[dim, :] += amp
        part = lattice_pass(ring, part, nturns=nturns)
        sh = part.shape
        partx = numpy.reshape(part[0, :], (sh[1], sh[3]))
        partxp = numpy.reshape(part[1, :], (sh[1], sh[3]))
        party = numpy.reshape(part[2, :], (sh[1], sh[3]))
        partyp = numpy.reshape(part[3, :], (sh[1], sh[3]))
        px = numpy.array([numpy.matmul(invx, [partx[i], partxp[i]])
                          for i in range(len(amp))])
        py = numpy.array([numpy.matmul(invy, [party[i], partyp[i]])
                          for i in range(len(amp))])
        return px[:, 0, :] - 1j*px[:, 1, :], py[:, 0, :] - 1j*py[:, 1, :]

    l0, q0, _, _ = linopt(ring, dp=dp)
    orbit = l0['closed_orbit']
    tunes = []

    if amp is not None:
        partx, party = _gen_part(ring, amp, dim, orbit, l0, nturns)
        qx = get_tunes_harmonic(partx, 'laskar')
        qy = get_tunes_harmonic(party, 'laskar')
        tunes = numpy.vstack((qx, qy)).T

    return numpy.array(tunes)


def detuning(ring, xm=0.3e-4, ym=0.3e-4, npoints=3, dp=0, window=1,
             nturns=512):
    """
    This function uses tunes_vs_amp to compute the tunes for
    the specified amplitudes. Then it fits this data and returns
    result for dQx/dx, dQy/dx, dQx/dy, dQy/dy
    """
    lindata0, _, _, _ = linopt(ring, dp=dp)
    gamma = (1 + lindata0.alpha * lindata0.alpha) / lindata0.beta

    x = numpy.linspace(-xm, xm, npoints)
    y = numpy.linspace(-ym, ym, npoints)
    x2 = x * x
    y2 = y * y

    q_dx = tunes_vs_amp(ring, amp=x, dim=0, dp=dp, window=window,
                        nturns=nturns)
    q_dy = tunes_vs_amp(ring, amp=y, dim=2, dp=dp, window=window,
                        nturns=nturns)

    fx = numpy.polyfit(x2, q_dx, 1)
    fy = numpy.polyfit(y2, q_dy, 1)

    q0 = [[fx[1, 0], fx[1, 1]],
          [fy[1, 0], fy[1, 1]]]
    q1 = [[2 * fx[0, 0] / gamma[0], 2 * fx[0, 1] / gamma[0]],
          [2 * fy[0, 0] / gamma[1], 2 * fy[0, 1] / gamma[1]]]

    return numpy.array(q0), numpy.array(q1), x, q_dx, y, q_dy


def chromaticity(ring, method='linopt', dpm=0.02, npoints=11, order=3, dp=0,
                 **kwargs):
    """
    This function uses computes the tunes to compute the tune for
    the specified momentum offsets. Then it fits this data and returns
    the chromaticity up to the given order (npoints>order)
    OUTPUT
        (fitx, fity), dpa, qz
    """
    if order == 0:
        return get_chrom(ring, dp=dp, **kwargs)
    elif order > npoints - 1:
        raise ValueError('order should be smaller than npoints-1')
    else:
        dpa = numpy.linspace(-dpm, dpm, npoints)
        qz = []
        for dpi in dpa:
            qz.append(get_tune(ring, method=method, dp=dp+dpi, remove_dc=True,
                      **kwargs))
        fit = numpy.polyfit(dpa, qz, order)[::-1]
        fitx = fit[:, 0]/factorial(numpy.arange(order + 1))
        fity = fit[:, 1]/factorial(numpy.arange(order + 1))
        return numpy.array([fitx, fity]), dpa, numpy.array(qz)


def gen_detuning_elem(ring, orbit=None):
    """
    Generates an element that for detuning with amplitude
    """
    if orbit is None:
        orbit, _ = find_orbit4(ring)
    [lindata0, tunes, xsi, lindata] = ring.linopt(dp=0,
                                                  get_chrom=True,
                                                  orbit=orbit)

    r0, r1, x, q_dx, y, q_dy = detuning(ring, xm=1.0e-4,
                                        ym=1.0e-4, npoints=3, dp=0)
    rp = ring.periodicity
    nonlin_elem = Element('NonLinear', PassMethod='DeltaQPass',
                          Betax=lindata0.beta[0], Betay=lindata0.beta[1],
                          Alphax=lindata0.alpha[0],
                          Alphay=lindata0.alpha[1],
                          Qpx=xsi[0], Qpy=xsi[1], A1=r1[0][0]*rp, A2=r1[0][1]*rp,
                          A3=r1[1][1]*rp, T1=-orbit, T2=orbit)
    return nonlin_elem



"""Text representation of a python AT lattice with each element represented by
its 'repr' string
"""
from __future__ import print_function
import sys
from os.path import abspath
import numpy
from at.lattice import Lattice
from at.load import register_format
from at.load.utils import element_from_string

__all__ = ['load_repr', 'save_repr']


def load_repr(filename, **kwargs):
    """Load a Lattice object stored as a repr-file

    PARAMETERS
        filename        name of the '.repr' file

    KEYWORDS
        name            Name of the lattice (default: taken from the file)
        energy          Energy of the lattice (default: taken from the file)
        periodicity     Number of periods (default: taken from the file)
        *               all other keywords will be set as Lattice attributes

    OUTPUT
        Lattice object
    """
    def elem_iterator(params, repr_file):
        with open(params.setdefault('repr_file', repr_file), 'rt') as file:
            # the 1st line is the dictionary of saved lattice parameters
            for k, v in eval(next(file)).items():
                params.setdefault(k, v)
            for line in file:
                yield element_from_string(line.strip())

    return Lattice(abspath(filename), iterator=elem_iterator, **kwargs)


def save_repr(ring, filename=None):
    """Save a Lattice object as a repr-file

    PARAMETERS
        ring            Lattice object
        filename=None   name of the '.repr' file. Default: output on sys.stdout
    """
    def save(file):
        print(repr(dict((k, v) for k, v in vars(ring).items()
                        if not k.startswith('_'))), file=file)
        for elem in ring:
            print(repr(elem), file=file)

    # Save the current options
    opts = numpy.get_printoptions()
    # Set options to print the full representation of float variables
    numpy.set_printoptions(formatter={'float_kind': repr})
    if filename is None:
        save(sys.stdout)
    else:
        with open(filename, 'wt') as reprfile:
            save(reprfile)
    # Restore the current options
    numpy.set_printoptions(**opts)


register_format('.repr', load_repr, save_repr,
                descr='Text representation of a python AT Lattice')



"""AT plotting functions"""
from at.lattice import Lattice
from at.tracking import lattice_pass
from at.physics import linopt
from at.plot import baseplot

# --------- Example 1 --------

# The specific data generating functions do not depend on any graphics-related
# function. So they could be located in any module, for instance
# in lattice/linear.py where the linopt function is.


# data generating function
def pldata_beta_disp(ring, refpts, **kwargs):
    """Generates data for plotting beta functions and dispersion"""

    # compute linear optics at the required locations
    data0, _, _, data = linopt(ring, refpts=refpts, get_chrom=True, **kwargs)

    # Extract the plot data
    s_pos = data['s_pos']
    betax = data['beta'][:, 0]
    betaz = data['beta'][:, 1]
    dispersion = data['dispersion'][:, 0]

    # Left axis definition
    left = (r'$\beta$ [m]', s_pos, [betax, betaz],
            [r'$\beta_x$', r'$\beta_z$'])
    # Right axis definition
    right = ('dispersion [m]', s_pos, [dispersion], ['dispersion'])
    return 'Optical functions', left, right


# Convenience function to make the call simpler
def plot_beta(ring, **kwargs):
    """
    Plot beta functions and dispersion

    PARAMETERS
        ring            Lattice object

    KEYWORDS
        dp=0.0          momentum deviation.
        orbit           avoids looking for the closed orbit if is already known
                        ((6,) array)
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to False
        ddp=1.0E-8      momentum deviation used for computation of
                        chromaticities and dispersion
        coupled=True    if False, simplify the calculations by assuming
                        no H/V coupling
    """
    return baseplot(ring, pldata_beta_disp, **kwargs)

# --------- Example 2 --------


# data generating function
def pldata_linear(ring, refpts, *keys, **kwargs):
    """data extraction function for plotting results of linopt"""

    class Lind(object):
        """helper class for lindata extraction"""
        lab2 = "xz"
        lab6 = ("x", "x'", "z", "z'", "l", "\\delta")
        id6 = "123456"
        params = dict(
            beta=(r'$\beta$ [m]', r'$\beta_{0}$', lab2),
            closed_orbit=('position [m]', r'${0}$', lab6),
            dispersion=('dispersion [m]', r'$\eta_{0}$', lab6),
            alpha=(r'$\alpha$', r'$\alpha_{0}$', lab2),
            mu=(r'Phase advance', r'$\mu_{0}$', lab2),
            gamma=('Gamma', 'Gamma', None),
            A=('A', r'$A_{{{0}{1}}}$', id6),
            B=('B', r'$B_{{{0}{1}}}$', id6),
            C=('C', r'$C_{{{0}{1}}}$', id6),
            m44=('Transfert', r'$T_{{{0},{1}}}$', id6)
        )

        @classmethod
        def extract(cls, lindata, key, *idx):
            def it(v):
                try:
                    return iter(v)
                except TypeError:
                    return iter([v])

            axis_title, fmt, convert = cls.params[key]
            indices = list(zip(*(it(i) for i in idx)))
            print(indices)
            datay = (lindata[key][(slice(None),) + ic] for ic in indices)
            labels = (fmt.format(*(convert[i] for i in ic)) for ic in indices)
            return axis_title, lindata['s_pos'], datay, labels

    title = kwargs.pop('title', 'Linear optics')
    data0, _, _, data = linopt(ring, refpts=refpts, get_chrom=True, **kwargs)
    return (title,) + tuple(Lind.extract(data, *key) for key in keys)


# Convenience function to make the call simpler
def plot_linear(ring, *keys, **kwargs):
    """
    axleft, axright = plot_linear(ring, left[, right], **keywords
    Plot linear optical functions returned by linopt

    PARAMETERS
        ring            Lattice object
        left            Left axis description as a tuple:
                        (key[, indices[, indices]])
                          key:        'beta', 'closed_orbit',...
                          indices:    integer, sequence of integers, or slice
                        The number if sequences of indices is data[key].ndim-1
                        The number of indices is the number of curves to plot.
                        All sequences must have the same length.

            Examples:
              ('beta', [0, 1])              beta_x, beta_z
              ('dispersion', 0)             eta_x
              ('closed_orbit'), [1, 3])     x', z'
              ('m44', 2, 2)                 T33
              ('m44', [0, 0], [0, 1])       T11, T12
              ('m44', 2, slice(4))          T31, T32, T33, T34
                                            as a single block
              ('m44', [2,2,2,2], [0,1,2,3]) T31, T32, T33, T34

        right           Right axis (optional)

    KEYWORDS
        title           Plot title, defaults to "Linear optics"
        dp=0.0          momentum deviation.
        orbit           avoids looking for the closed orbit if is already known
                        ((6,) array)
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to False
        ddp=1.0E-8      momentum deviation used for computation of
                        chromaticities and dispersion
        coupled=True    if False, simplify the calculations by assuming
                        no H/V coupling
    """
    return baseplot(ring, pldata_linear, *keys, **kwargs)


# --------- Example 3 --------

# Here the data generating function is embedded in the convenience function
def plot_trajectory(ring, r_in, nturns=1, **kwargs):
    """
    plot a particle's trajectory

    PARAMETERS
        ring            Lattice object
        r_in            6xN array: input coordinates of N particles
        nturns=1        Number of turns

    KEYWORDS
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to False
    """
    # noinspection PyShadowingNames
    def pldata_trajectory(ring, refpts, r_in, nturns=1, **kwargs):
        r_out = lattice_pass(ring, r_in, refpts=refpts, nturns=nturns,
                             **kwargs)
        s_pos = ring.get_s_pos(refpts)
        particles = range(r_out.shape[1])
        xx = [r_out[0, i, :, :] for i in particles]
        zz = [r_out[2, i, :, :] for i in particles]
        xlabels = [r'$x_{0}$'.format(i) for i in particles]
        zlabels = [r'$z_{0}$'.format(i) for i in particles]
        left = ('position [m]', s_pos, xx+zz, xlabels+zlabels)
        return 'Trajectory', left

    return baseplot(ring, pldata_trajectory, r_in, nturns=nturns, **kwargs)


Lattice.plot_beta = plot_beta
Lattice.plot_trajectory = plot_trajectory



"""Simple tests to make sure integrators load in Python."""
import numpy
import pytest

from at import atpass, elements
from at.lattice import Lattice


def test_exact_hamiltonian_pass(rin):
    drift = elements.Multipole('m1', 1, [0, 0, 0, 0], [0, 0, 0, 0])
    drift.Type = 0
    drift.PassMethod = 'ExactHamiltonianPass'
    drift.BendingAngle = 0
    l = Lattice([drift], name='lat', energy=3e9)
    atpass(l, rin, 1)


def test_exact_hamiltonian_pass_with_dls_dipole(rin):
    bend = elements.Multipole('rb', 0.15, [0, 0, 0, 0], [-0.0116333, 3.786786, 0, 0])
    bend.Type = 1
    bend.PassMethod = 'ExactHamiltonianPass'
    bend.BendingAngle = -0.001745
    bend.Energy = 3.5e9
    bend.MaxOrder = 3
    l = Lattice([bend], name='lat', energy=3.5e9)
    atpass(l, rin, 1)
    # Results from Matlab
    expected = numpy.array([9.23965e-9, 1.22319e-5, 0, 0, 0, -4.8100e-10]).reshape(6,1)
    numpy.testing.assert_allclose(rin, expected, rtol=1e-5, atol=1e-6)


@pytest.mark.parametrize('passmethod', ('GWigSymplecticPass', 'GWigSymplecticRadPass'))
def test_gwig_symplectic_pass(rin, passmethod):
    # Parameters copied from one of the Diamond wigglers.
    wiggler = elements.Wiggler('w', 1.15, 0.05, 0.8, 3e9)
    wiggler.PassMethod = passmethod
    l = Lattice([wiggler], name='lat', energy=3e9)
    atpass(l, rin, 1)


def test_bndstrmpole_symplectic_4_pass(rin):
    bend = elements.Dipole('b', 1.0)
    bend.PassMethod = 'BndStrMPoleSymplectic4Pass'
    l = Lattice([bend], name='lat', energy=3e9)
    atpass(l, rin, 1)



"""
Functions relating to particle generation
"""
import numpy
from at.physics import ohmi_envelope
from at.lattice.lattice_object import Lattice

__all__ = ['beam', 'sigma_matrix']


def sigma_matrix(argin):
    """
    Calculate the correlation matrix to be used for particle generation

    PARAMETERS
        argin           Lattice object or list of twiss parameters.
                        Depending on the length of the list the function
                        will do different things:
                        [espread, blength] 
                            computes 2x2 longitudinal sigma matrix
                        [beta, alpha, emit] 
                            computes 2x2 transverse sigma matrix
                        [betax, alphax, emitx, betay, alphay, emity] 
                            computes the 4x4 sigma matrix for horizontal 
                            and vertical.
                        [betax, alphax, emitx, betay, alphay, emity, espread, 
                            blength] 
                            computes the 6x6 sigma matrix.

    KEYWORDS
        twiss=False     Flag which states whether the input is a
                        lattice object or list of twiss parameters

    OUTPUT
        sigma_matrix    correlation matrix (either 2x2, 4x4 or 6x6)

    """
    if isinstance(argin, Lattice):
        argin = argin.radiation_on(copy=True)

        emit0, beamdata, emit = ohmi_envelope(argin, refpts=[0])
        sig_matrix = emit.r66[0]


    else:
        if len(argin) == 2:
            [espread, blength] = argin
            sig_matrix = numpy.array([
                            [espread*espread, 0],
                            [0, blength*blength]
                                     ])

        elif len(argin) == 3:
            [bx, ax, epsx] = argin
            sig_matrix = epsx * numpy.array([
                                    [bx, -ax],
                                    [-ax, (1 + ax * ax)/bx]
                                            ])

        elif len(argin) == 6:
            [bx, ax, epsx, by, ay, epsy] = argin
            sig_matrix = numpy.block([
                            [sigma_matrix([bx, ax, epsx]),
                                numpy.zeros((2, 2))],
                            [numpy.zeros((2, 2)),
                                sigma_matrix([by, ay, epsy])]
                                    ])

        elif len(argin) == 8:
            [bx, ax, epsx, by, ay, epsy, espread, blength] = argin
            sig_matrix = numpy.block([
                            [sigma_matrix([bx, ax, epsx]),
                                numpy.zeros((2, 4))],
                            [numpy.zeros((2, 2)),
                                sigma_matrix([by, ay, epsy]),
                                numpy.zeros((2, 2))],
                            [numpy.zeros((2, 4)),
                                sigma_matrix([espread, blength])]
                                    ])
        else:
            raise AttributeError('Wrong number of inputs provided')

    return sig_matrix


def beam(np, sigma, orbit=None):
    """
    Generates an array of random particles according to the given sigma
    matrix

    PARAMETERS
        np              Number of particles
        sigma           sigma_matrix as calculated by at.sigma_matrix

    KEYWORDS
        orbit=None      An orbit can be provided to give a center of 
                        mass offset to the distribution

    OUTPUT
        particle_dist   a matrix of shape (M, np) where M is shape of
                        sigma matrix
    """
    v = numpy.random.normal(size=(sigma.shape[0], np))

    try:
        lmat = numpy.linalg.cholesky(sigma)

    except numpy.linalg.LinAlgError:
        row_idx = numpy.array([0, 1, 4, 5])
        a1 = numpy.linalg.cholesky(sigma[row_idx[:, None], row_idx])
        a = numpy.block([[a1, numpy.zeros((4, 2))], [numpy.zeros((2, 6))]])
        row_idx = numpy.array([0, 1, 4, 5, 2, 3])
        lmat = a[row_idx[:, None], row_idx]

    particle_dist = numpy.squeeze(numpy.dot(lmat, v))

    if orbit is not None:
        if (not isinstance(orbit, (numpy.ndarray, list)) or
                len(orbit) != sigma.shape[0]):
            raise AttributeError('orbit should be a list or array' +
                                 ' with a length the same as sigma' +
                                 ' matrix')
        else:
            for i, orb in enumerate(orbit):
                particle_dist[i, :] += orb

    return particle_dist



"""
Load lattices from Matlab files.
"""
from __future__ import print_function
import sys
from os.path import abspath
from warnings import warn
import scipy.io
import numpy
from at.lattice import elements, Lattice, AtWarning, params_filter
from at.load import register_format
from at.load.utils import element_from_dict, element_from_m, RingParam
from at.load.utils import element_to_dict, element_to_m

__all__ = ['ringparam_filter', 'load_mat', 'save_mat', 'load_m', 'save_m',
           'load_var']

_param_to_lattice = {'Energy': 'energy', 'Periodicity': 'periodicity',
                     'FamName': 'name'}
_param_ignore = {'PassMethod', 'Length'}

# Python to Matlab attribute translation
_matattr_map = dict(((v, k) for k, v in _param_to_lattice.items()))


def matfile_generator(params, mat_file):
    """
    run through matlab cells and generate AT elements

    KEYWORDS
        mat_file        name of the .mat file
        mat_key         name of the Matlab variable containing the lattice.
                        Default: Matlab variable name if there is only one,
                        otherwise 'RING'
        check=True      if False, skip the coherence tests
        quiet=False     If True, suppress the warning for non-standard classes
    """
    m = scipy.io.loadmat(params.setdefault('mat_file', mat_file))
    matvars = [varname for varname in m if not varname.startswith('__')]
    default_key = matvars[0] if (len(matvars) == 1) else 'RING'
    key = params.setdefault('mat_key', default_key)
    check = params.pop('check', True)
    quiet = params.pop('quiet', False)
    cell_array = m[key].flat
    for index, mat_elem in enumerate(cell_array):
        element_array = mat_elem[0, 0]
        kwargs = {}
        for field_name in element_array.dtype.fields:
            # Remove any surplus dimensions in arrays.
            data = numpy.squeeze(element_array[field_name])
            # Convert strings in arrays back to strings.
            if data.dtype.type is numpy.unicode_:
                data = str(data)
            kwargs[field_name] = data

        yield element_from_dict(kwargs, index=index, check=check, quiet=quiet)


def ringparam_filter(params, elem_iterator, *args):
    """"
    Run through all elements, process and optionally removes RingParam elements

    KEYWORDS
        keep_all=False  if True, keep RingParam elem_iterator as Markers
    """
    keep_all = params.pop('keep_all', False)
    ringparams = []
    radiate = False
    for elem in elem_iterator(params, *args):
        if (elem.PassMethod.endswith('RadPass') or
                elem.PassMethod.endswith('CavityPass')):
            radiate = True
        if isinstance(elem, RingParam):
            ringparams.append(elem)
            for k, v in elem.items():
                if k not in _param_ignore:
                    params.setdefault(_param_to_lattice.get(k, k.lower()), v)
            if keep_all:
                pars = vars(elem).copy()
                name = pars.pop('FamName')
                yield elements.Marker(name, **pars)
        else:
            yield elem
    params['_radiation'] = radiate

    if len(ringparams) > 1:
        warn(AtWarning('More than 1 RingParam element, the 1st one is used'))


def load_mat(filename, **kwargs):
    """Create a lattice object from a matmab mat-file

    PARAMETERS
        filename        name of a '.mat' file

    KEYWORDS
        mat_key         name of the Matlab variable containing the lattice.
                        Default: Matlab variable name if there is only one,
                        otherwise 'RING'
        check=True      if False, skip the coherence tests
        quiet=False     If True, suppress the warning for non-standard classes
        keep_all=False  if True, keep RingParam elements as Markers
        name            Name of the lattice
                        (default: taken from the lattice, or '')
        energy          Energy of the lattice
                        (default: taken from the elements)
        periodicity     Number of periods
                        (default: taken from the elements, or 1)
        *               all other keywords will be set as Lattice attributes

    OUTPUT
        Lattice object
    """
    if 'key' in kwargs:  # process the deprecated 'key' keyword
        kwargs.setdefault('mat_key', kwargs.pop('key'))
    return Lattice(ringparam_filter, matfile_generator, abspath(filename),
                   iterator=params_filter, **kwargs)


def mfile_generator(params, m_file):
    """Run through all lines of a m-file and generates AT elements"""
    with open(params.setdefault('m_file', m_file), 'rt') as file:
        _ = next(file)  # Matlab function definition
        _ = next(file)  # Cell array opening
        for lineno, line in enumerate(file):
            if line.startswith('};'):
                break
            try:
                elem = element_from_m(line)
            except ValueError:
                warn(AtWarning('Invalid line {0} skipped.'.format(lineno)))
                continue
            except KeyError:
                warn(AtWarning('Line {0}: Unknown class'))
                continue
            else:
                yield elem


def load_m(filename, **kwargs):
    """Create a lattice object from a matlab m-file

    PARAMETERS
        filename        name of a '.m' file

    KEYWORDS
        keep_all=False  if True, keep RingParam elements as Markers
        name            Name of the lattice
                        (default: taken from the elements, or '')
        energy          Energy of the lattice
                        (default: taken from the elements)
        periodicity     Number of periods
                        (default: taken from the elements, or 1)
        *               all other keywords will be set as Lattice attributes

    OUTPUT
        Lattice object
    """
    return Lattice(ringparam_filter, mfile_generator, abspath(filename),
                   iterator=params_filter, **kwargs)


def load_var(matlat, **kwargs):
    """Create a lattice from a Matlab cell array"""

    # noinspection PyUnusedLocal
    def var_generator(params, latt):
        for elem in latt:
            yield element_from_dict(elem)

    return Lattice(ringparam_filter, var_generator, matlat,
                   iterator=params_filter, **kwargs)


def matlab_ring(ring):
    """Prepend a RingParam element to a lattice"""
    dct = dict((_matattr_map.get(k, k.title()), v)
               for k, v in vars(ring).items() if not k.startswith('_'))
    famname = dct.pop('FamName')
    energy = dct.pop('Energy')
    yield RingParam(famname, energy, **dct)
    for elem in ring:
        yield elem


def save_mat(ring, filename, mat_key='RING'):
    """Save a Lattice object as a Matlab mat-file

    PARAMETERS
        ring            Lattice object
        filename        name of the '.mat' file

    KEYWORDS
        mat_key='RING'  Name of the Matlab variable representing the lattice
    """
    lring = tuple((element_to_dict(elem),) for elem in matlab_ring(ring))
    scipy.io.savemat(filename, {mat_key: lring})


def save_m(ring, filename=None):
    """Save a lattice as a Matlab m-file

    PARAMETERS
        ring            Lattice object
        filename=None   name of the '.m' file. Default: output on sys.stdout
    """

    def save(file):
        print('function ring = {0}()'.format(ring.name), file=file)
        print('ring = {...', file=file)
        for elem in matlab_ring(ring):
            print(element_to_m(elem), file=file)
        print('};', file=file)
        print('end', file=file)

    if filename is None:
        save(sys.stdout)
    else:
        with open(filename, 'wt') as mfile:
            save(mfile)


register_format('.mat', load_mat, save_mat, descr='Matlab binary mat-file')
register_format('.m', load_m, save_m, descr='Matlab text m-file')



"""Lattice object

The methods implemented in this module are internal to the 'lattice' package.
This is necessary to ensure that the 'lattice' package is independent
from other AT packages.

Other Lattice methods are implemented in other AT packages and are available
as soon as the package is imported. The 'tracking' and 'physics' packages are
automatically imported.

As an example, see the at.physics.orbit module
"""
import sys
import copy
import numpy
import math
import itertools
from warnings import warn
from at.lattice import AtError, AtWarning
from at.lattice import elements, get_s_pos, get_elements, uint32_refpts

TWO_PI_ERROR = 1.E-4

__all__ = ['Lattice', 'type_filter', 'params_filter', 'lattice_filter',
           'no_filter']


class Lattice(list):
    """Lattice object
    An AT lattice is a sequence of AT elements.
    A Lattice accepts extended indexing (as a numpy ndarray).

    Lattice attributes;
        name        Name of the lattice
        energy      Particle energy
        periodicity Number of super-periods to describe the full ring

    Lattice(elems, **params)     Create a new lattice object

        INPUT
            elems:          any iterable of AT elements

        KEYWORDS
            iterator=None   Custom iterator (see below)
            scan=False      Scan elements, looking for energy and periodicity
            name            Name of the lattice
            energy          Energy of the lattice
            periodicity     Number of periods
            *               all other keywords will be set as Lattice attributes

    To reduce the inter-package dependencies, some methods of the
    lattice object are defined in other AT packages, in the module where
    the underlying function is implemented.

    Custom iterators:
    Instead of running through 'elems', the Lattice constructor can use one
    or several custom iterators.

    Lattice(*args, iterator=it, **params)

        The iterator "it" is called as "it(params, *args)" and must return an
            iterator over AT elements for building the lattice. It must also
            fill the "params" dictionary used to set the Lattice attributes.

            params is the dictionary of lattice parameters. It is initialised
                   with the keywords of the lattice constructor. It can be
                   modified by the custom iterator to add or remove parameters.
                   Finally, all remaining parameters will be set as Lattice
                   attributes.
            *args  all positional arguments of the Lattice constructor are sent
                   to the custom iterator.

        An iterator can be:
            - a "generator" which yields elements from scratch.
              Examples: a list, or a file iterator,
            - a "filter" which runs through an input iterator, processes each
              element, possibly adds parameters to the params dictionary
              and yields the processed elements.

        Example of chaining iterators (taken from "load_mat"):

        Lattice(ringparam_filter, matfile_generator, filename
                iterator=params_filter, **params)

        matfile_generator(params, filename)
            opens filename and generates AT elements for each cell of the
            Matlab cell array representing the lattice,

        ringparam_filter(params, matfile_generator, *args)
            runs through matfile_generator(params, *args), looks for RingParam
            elements, fills params with their information and discards them,

        params_filter(params, ringparam_filter, *args)
            runs through ringparam_filter(params, *args), looks for energy and
            periodicity if not yet defined.
    """
    _1st_attributes = ('name', 'energy', 'periodicity')

    def __init__(self, *args, **kwargs):
        """Lattice constructor"""

        iterator = kwargs.pop('iterator', None)
        scan = kwargs.pop('scan', False)
        for key in list(k for k in kwargs.keys() if k.startswith('_')):
            kwargs.pop(key)
        if iterator is None:
            elem_list, = args or [[]]  # accept 0 or 1 argument
            if isinstance(elem_list, Lattice):
                elems = lattice_filter(kwargs, elem_list)
            elif scan:
                elems = params_filter(kwargs, type_filter, elem_list)
            else:
                elems = type_filter(kwargs, elem_list)
        else:
            elems = iterator(kwargs, *args)

        super(Lattice, self).__init__(elems)

        # set default values
        kwargs.setdefault('name', '')
        kwargs.setdefault('periodicity', 1)
        if 'energy' not in kwargs:
            raise AtError('Lattice energy is not defined')
        # set attributes
        self.update(kwargs)

    def __getitem__(self, key):
        try:
            return super(Lattice, self).__getitem__(key.__index__())
        except (AttributeError, TypeError):
            return Lattice(self.iterator(key), iterator=no_filter, **vars(self))

    if sys.version_info < (3, 0):
        # This won't be defined if version is at least 3.0
        # Called for slices with step != 1
        # noinspection PyTypeChecker
        def __getslice__(self, i, j):
            return self.__getitem__(slice(i, j))

    def __setitem__(self, key, values):
        try:
            super(Lattice, self).__setitem__(key, values)
        except TypeError:
            key = uint32_refpts(key, len(self))
            for i, v in zip(*numpy.broadcast_arrays(key, values)):
                super(Lattice, self).__setitem__(i, v)

    def __delitem__(self, key):
        try:
            super(Lattice, self).__delitem__(key)
        except TypeError:
            key = uint32_refpts(key, len(self))
            for i in reversed(key):
                super(Lattice, self).__delitem__(i)

    def __repr__(self):
        attrs = vars(self).copy()
        k1 = [(k, attrs.pop(k)) for k in Lattice._1st_attributes]
        k2 = [(k, v) for k, v in attrs.items() if not k.startswith('_')]
        keys = ', '.join('{0}={1!r}'.format(k, v) for k, v in (k1 + k2))
        return 'Lattice({0}, {1})'.format(super(Lattice, self).__repr__(), keys)

    def __str__(self):
        attrs = vars(self).copy()
        k1 = [(k, attrs.pop(k)) for k in Lattice._1st_attributes]
        k2 = [(k, v) for k, v in attrs.items() if not k.startswith('_')]
        keys = ', '.join('{0}={1!r}'.format(k, v) for k, v in (k1 + k2))
        return 'Lattice(<{0} elements>, {1})'.format(len(self), keys)

    def __add__(self, elems):
        return Lattice(itertools.chain(self, elems), **vars(self))

    def __mul__(self, times):
        return Lattice(itertools.chain(*itertools.repeat(self, times)),
                       **vars(self))

    def iterator(self, key):
        """Iterates over the indices selected by a slice or an array"""
        if isinstance(key, slice):
            indices = key.indices(len(self))
            rg = range(*indices)
        else:
            rg = uint32_refpts(key, len(self))
        return (super(Lattice, self).__getitem__(i) for i in rg)

    def update(self, *args, **kwargs):
        """Update the element attributes with the given arguments

        update(**kwargs)
        update(mapping, **kwargs)
        update(iterable, **kwargs)
        """
        attrs = dict(*args, **kwargs)
        for (key, value) in attrs.items():
            setattr(self, key, value)

    def copy(self):
        """Return a shallow copy"""
        return Lattice(self)

    def deepcopy(self):
        """Return a deep copy"""
        return copy.deepcopy(self)

    def slice(self, size=None, slices=1):
        """Create a new lattice by slicing the range of interest into small
        elements

        KEYWORDS
            size=None       Length of a slice. Default: computed from the
                            range and number of points:
                                    sx = (s_max-s_min)/slices.
            slices=1        Number of slices in the specified range. Ignored if
                            size is specified. Default: no slicing

        RETURN
            New Lattice object
       """

        def slice_iter(ibeg, iend):
            for elem in self[:ibeg]:
                yield elem
            for elem in self[ibeg:iend]:
                nslices = int(math.ceil(elem.Length / size))
                if nslices > 1:
                    frac = numpy.ones(nslices) / nslices
                    for el in elem.divide(frac):
                        yield el
                else:
                    yield elem
            for elem in self[iend:]:
                yield elem

        s_range = self.s_range
        if size is None:
            smin, smax = s_range
            size = (smax - smin) / slices

        i1 = self._i_range[0]
        i2 = self._i_range[-1]
        return Lattice(slice_iter(i1, i2), s_range=s_range, **vars(self))

    @property
    def s_range(self):
        """Range of interest. 'None' means the full cell."""
        try:
            return self._s_range
        except AttributeError:
            self.s_range = None
            return self._s_range

    # noinspection PyAttributeOutsideInit
    @s_range.setter
    def s_range(self, value):
        spos = self.get_s_pos(range(len(self) + 1))
        if value is None:
            value = (0.0, spos[-1])
        ok = numpy.flatnonzero(
            numpy.logical_and(spos > value[0], spos < value[1]))
        if len(ok) > 0:
            i1 = max(ok[0] - 1, 0)
            i2 = min(ok[-1] + 1, len(self))
            self._i_range = range(i1, i2 + 1)
        else:
            self._i_range = range(0, 0)
        self._s_range = value

    @property
    def i_range(self):
        """Range of elements inside the range of interest"""
        try:
            i_range = self._i_range
        except AttributeError:
            self.s_range = None
            i_range = self._i_range
        return uint32_refpts(i_range, len(self))

    @property
    def circumference(self):
        """Ring circumference"""
        return self.periodicity * self.get_s_pos(len(self))[0]

    @property
    def voltage(self):
        """Total accelerating voltage"""
        volts = [elem.Voltage for elem in self if
                 isinstance(elem, elements.RFCavity)]
        return self.periodicity * sum(volts)

    @property
    def harmonic_number(self):
        """Harmonic number"""
        harms = [elem.HarmNumber for elem in self if
                 isinstance(elem, elements.RFCavity)]
        return self.periodicity * harms[0] if harms else None

    @property
    def radiation(self):
        """If True, at least one element modifies the beam energy"""
        try:
            return self._radiation
        except AttributeError:
            radiate = False
            for elem in self:
                if (elem.PassMethod.endswith('RadPass') or
                        elem.PassMethod.endswith('CavityPass')):
                    radiate = True
                    break
            # noinspection PyAttributeOutsideInit
            self._radiation = radiate
            return radiate

    # noinspection PyShadowingNames
    def modify_elements(self, elem_modify, copy=True):
        """Modify selected elements, in-place or in a lattice copy

        PARAMETERS
            elem_modify         element selection function.

            If elem_modify(elem) returns None, the element is unchanged.
            Otherwise, elem_modify(elem) must return a dictionary of
            attribute name and values, to be set to elem.

        RETURNS
            New lattice if copy == True
            None if copy == False

        KEYWORDS
            copy=True           Return a shallow copy of the lattice. Only the
                                modified attributes are replaced. Otherwise
                                modify the lattice in-place.
        """

        def lattice_modify():
            """Modifies the Lattice with elements modified by elem_modify"""
            radiate = False
            for elem in self:
                attrs = elem_modify(elem)
                if attrs is not None:
                    elem.update(attrs)
                if (elem.PassMethod.endswith('RadPass') or
                        elem.PassMethod.endswith('CavityPass')):
                    radiate = True
            self._radiation = radiate

        def lattice_copy(params):
            """Custom iterator for the creation of a new lattice"""
            radiate = False
            for elem in lattice_filter(params, self):
                attrs = elem_modify(elem)
                if attrs is not None:
                    elem = elem.copy()
                    elem.update(attrs)
                if (elem.PassMethod.endswith('RadPass') or
                        elem.PassMethod.endswith('CavityPass')):
                    radiate = True
                yield elem
            params['_radiation'] = radiate

        if copy:
            return Lattice(iterator=lattice_copy)
        else:
            lattice_modify()

    @staticmethod
    def _radiation_attrs(cavity_func, dipole_func,
                         quadrupole_func, wiggler_func,
                         sextupole_func, octupole_func):
        """Create a function returning the modified attributes"""

        def elem_func(elem):

            def isdipole(el):
                return isinstance(el, elements.Dipole) and (
                        el.BendingAngle != 0.0)

            if isinstance(elem, elements.RFCavity):
                return cavity_func(elem)
            elif isdipole(elem):
                return dipole_func(elem)
            elif isinstance(elem, elements.Quadrupole):
                return quadrupole_func(elem)
            elif isinstance(elem, elements.Wiggler):
                return wiggler_func(elem)
            elif isinstance(elem, elements.Sextupole):
                return sextupole_func(elem)
            elif isinstance(elem, elements.Octupole):
                return octupole_func(elem)
            else:
                return None

        return elem_func

    # noinspection PyShadowingNames
    def radiation_on(self, cavity_pass='CavityPass', dipole_pass='auto',
                     quadrupole_pass=None, wiggler_pass='auto',
                     sextupole_pass=None, octupole_pass=None, copy=False):
        """
        Turn acceleration and radiation on and return the lattice

        KEYWORDS
            cavity_pass='CavityPass'    PassMethod set on cavities
            dipole_pass='auto'          PassMethod set on dipoles
            quadrupole_pass=None        PassMethod set on quadrupoles
            wiggler_pass='auto'         PassMethod set on wigglers
            copy=False                  Return a shallow copy of the lattice and
                                        replace only the modified attributes
                                        Otherwise modify the lattice in-place.

            For PassMethod names, the convention is:
                None            no change
                'auto'          replace *Pass by *RadPass
                anything else   set as the new PassMethod
        """

        def repfunc(pass_method):
            if pass_method is None:
                # noinspection PyUnusedLocal
                def ff(elem):
                    return None
            elif pass_method == 'auto':
                def ff(elem):
                    if not elem.PassMethod.endswith('RadPass'):
                        pass_m = ''.join((elem.PassMethod[:-4], 'RadPass'))
                        return {'PassMethod': pass_m,
                                'Energy': self.energy}
                    else:
                        return None
            else:
                def ff(elem):
                    if elem.PassMethod != pass_method:
                        return {'PassMethod': pass_method,
                                'Energy': self.energy}
                    else:
                        return None
            return ff

        elem_func = self._radiation_attrs(repfunc(cavity_pass),
                                          repfunc(dipole_pass),
                                          repfunc(quadrupole_pass),
                                          repfunc(wiggler_pass),
                                          repfunc(sextupole_pass),
                                          repfunc(octupole_pass))
        return self.modify_elements(elem_func, copy=copy)

    # noinspection PyShadowingNames
    def radiation_off(self, cavity_pass='auto', dipole_pass='auto',
                      quadrupole_pass='auto', wiggler_pass='auto',
                      sextupole_pass='auto', octupole_pass='auto', copy=False):
        """
        Turn acceleration and radiation off and return the lattice

        KEYWORDS
            cavity_pass='IdentityPass'  PassMethod set on cavities
            dipole_pass='auto'          PassMethod set on dipoles
            quadrupole_pass=None        PassMethod set on quadrupoles
            wiggler_pass='auto'         PassMethod set on wigglers
            copy=False                  Return a shallow copy of the lattice and
                                        replace only the modified attributes
                                        Otherwise modify the lattice in-place.

            For PassMethod names, the convention is:
                None            no change
                'auto'          replace *RadPass by *Pass
                anything else   set as it is
        """

        def auto_cavity_pass(elem):
            newpass = 'IdentityPass' if elem.Length == 0 else 'DriftPass'
            if elem.PassMethod != newpass:
                return {'PassMethod': newpass}
            else:
                return None

        def auto_multipole_pass(elem):
            if elem.PassMethod.endswith('RadPass'):
                newpass = ''.join((elem.PassMethod[:-7], 'Pass'))
                return {'PassMethod': newpass}
            else:
                return None

        def repfunc(pass_method, auto_method):
            if pass_method is None:
                # noinspection PyUnusedLocal
                def ff(elem):
                    return None
            elif pass_method == 'auto':
                ff = auto_method
            else:
                def ff(elem):
                    if elem.PassMethod != pass_method:
                        return {'PassMethod': pass_method}
                    else:
                        return None
            return ff

        elem_func = self._radiation_attrs(
            repfunc(cavity_pass, auto_cavity_pass),
            repfunc(dipole_pass, auto_multipole_pass),
            repfunc(quadrupole_pass, auto_multipole_pass),
            repfunc(wiggler_pass, auto_multipole_pass),
            repfunc(sextupole_pass, auto_multipole_pass),
            repfunc(octupole_pass, auto_multipole_pass))
        return self.modify_elements(elem_func, copy=copy)

    def sbreak(self, break_s, break_elems=None):
        """Insert elements at selected locations in the lattice

        PARAMETERS
            break_s:        location or array of locations of breakpoints
            break_elems:    elements to be inserted at breakpoints (array of
                            elements as long as break_s or single element
                            duplicated as necessary). Default: Marker('sbreak')
        RETURNS
            A new lattice with new elements inserted at breakpoints
        """

        def sbreak_iterator(elems, insertions):
            """Iterate over elements and breaks where necessary"""

            def next_mk():
                """Extract the next element to insert"""
                try:
                    return next(insertions)
                except StopIteration:
                    return sys.float_info.max, None

            s_end = 0.0
            # get the 1st insertion
            smk, mk = next_mk()
            # skip all insertions at negative break_s, if any
            while smk < s_end:
                smk, mk = next_mk()

            for elem in elems:
                s_end += elem.Length
                # loop over all insertions within the element
                while smk < s_end:
                    frac = (s_end - smk) / elem.Length
                    if frac < 1.0:  # breakpoint is within the element
                        el0, elem = elem.divide([1.0 - frac, frac])
                        yield el0
                    yield mk
                    smk, mk = next_mk()
                yield elem

        # set default insertion
        if break_elems is None:
            break_elems = elements.Marker('sbreak')
        break_elems = numpy.reshape(break_elems, -1)
        # Check element lengths
        if not all(e.Length==0 for e in break_elems):
            warn(AtWarning(
                 "Inserting elements with length!=0 may change the lattice"))
        # broadcast break_s and break_elems to arrays of same size
        # and create an iterator over the elements to be inserted
        iter_mk = zip(*numpy.broadcast_arrays(break_s, break_elems))

        return Lattice(sbreak_iterator(self, iter_mk), **vars(self))


def lattice_filter(params, elems):
    """Copy lattice parameters an run through all lattice elements"""
    for key, value in vars(elems).items():
        params.setdefault(key, value)
    return iter(elems)


# noinspection PyUnusedLocal
def no_filter(params, elems):
    """Run through all elements without any check"""
    return iter(elems)


def type_filter(params, elems):
    """Run through all elements and check element validity.
    Analyses elements for radiation state
    """
    radiate = False
    for idx, elem in enumerate(elems):
        if isinstance(elem, elements.Element):
            if (elem.PassMethod.endswith('RadPass') or
                    elem.PassMethod.endswith('CavityPass')):
                radiate = True
            yield elem
        else:
            warn(AtWarning('item {0} ({1}) is not an AT element: '
                           'ignored'.format(idx, elem)))
    params['_radiation'] = radiate


def params_filter(params, elem_iterator, *args):
    """Run through all elements, looking for energy and periodicity.
    Remove the Energy attribute of non-radiating elements

    energy is taken from:
        1) The params dictionary
        2) Radiative elements (Cavity, RingParam, Radiating magnets)
        3) Any other element

    periodicity is taken from:
        1) The params dictionary
        2) Sum of the bending angles of magnets
    """
    rad_energies = []
    el_energies = []
    thetas = []

    for idx, elem in enumerate(elem_iterator(params, *args)):
        if (isinstance(elem, (elements.RFCavity, elements.Wiggler)) or
                elem.PassMethod.endswith('RadPass')):
            rad_energies.append(elem.Energy)
            try:
                elem.Energy = params['energy']
            except KeyError:
                params['energy'] = elem.Energy
        elif hasattr(elem, 'Energy'):
            el_energies.append(elem.Energy)
            del elem.Energy
        if isinstance(elem, elements.Dipole):
            thetas.append(elem.BendingAngle)
        yield elem

    energies = rad_energies or el_energies

    if 'energy' not in params and energies:
        # Guess energy from the Energy attribute of the elements
        params['energy'] = max(energies)

    if energies and min(energies) < max(energies):
        warn(AtWarning('Inconsistent energy values, '
                       '"energy" set to {0}'.format(params['energy'])))

    if 'periodicity' not in params:
        # Guess periodicity from the bending angles of the lattice
        try:
            nbp = 2.0 * numpy.pi / sum(thetas)
        except ZeroDivisionError:
            periodicity = 1
            warn(AtWarning('No bending in the cell, set "Periodicity" to 1'))
        else:
            periodicity = int(round(nbp))
            if abs(periodicity - nbp) > TWO_PI_ERROR:
                warn(AtWarning('Non-integer number of cells: '
                               '{0} -> {1}'.format(nbp, periodicity)))
        params['periodicity'] = periodicity


Lattice.get_elements = get_elements
Lattice.get_s_pos = get_s_pos



import glob
from io import open
import os
from os.path import abspath, basename, dirname, exists, join, splitext
import sys
import shutil
from setuptools import setup, Extension, find_packages
try:
    import numpy
except ImportError:
    print('\npyAT requires numpy. '
          'Please install numpy: "pip install numpy"\n')
    sys.exit()


here = abspath(dirname(__file__))
macros = [('PYAT', None)]

cflags = []

if not sys.platform.startswith('win32'):
    cflags += ['-Wno-unused-function']


# Get the long description from the README file
with open(join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


# It is easier to copy the integrator files into a directory inside pyat
# for packaging. However, we cannot always rely on this directory being
# inside a clone of at and having the integrator sources available, and
# this file is executed each time any setup.py command is run.
# It appears that only copying the files when they are available is
# sufficient.
at_source = abspath(join(here, 'at.c'))
integrator_src_orig = abspath(join(here, '..', 'atintegrators'))
integrator_src = abspath(join(here, 'integrator-src'))
diffmatrix_source = abspath(
    join(here, '..', 'atmat', 'atphysics', 'Radiation')
)

if exists(integrator_src_orig):
    # Copy files into pyat for distribution.
    source_files = glob.glob(join(integrator_src_orig, '*.[ch]'))
    source_files.extend(glob.glob(join(integrator_src_orig, '*.cc')))
    source_files.extend(
        glob.glob(join(diffmatrix_source, 'findmpoleraddiffmatrix.c'))
    )
    if not exists(integrator_src):
        os.makedirs(integrator_src)
    for f in source_files:
        shutil.copy2(f, integrator_src)

pass_methods = glob.glob(join(integrator_src, '*Pass.c'))
pass_methods.extend(glob.glob(join(integrator_src, '*Pass.cc')))
diffmatrix_method = join(integrator_src, 'findmpoleraddiffmatrix.c')


def integrator_ext(pass_method):
    name, _ = splitext(basename(pass_method))
    name = ".".join(('at', 'integrators', name))
    return Extension(
        name=name,
        sources=[pass_method],
        include_dirs=[numpy.get_include(), integrator_src, diffmatrix_source],
        define_macros=macros,
        extra_compile_args=cflags
    )


at = Extension(
    'at.tracking.atpass',
    sources=[at_source],
    define_macros=macros,
    include_dirs=[numpy.get_include(), integrator_src, diffmatrix_source],
    extra_compile_args=cflags
)

diffmatrix = Extension(
    name='at.physics.diffmatrix',
    sources=[diffmatrix_method],
    include_dirs=[numpy.get_include(), integrator_src, diffmatrix_source],
    define_macros=macros,
    extra_compile_args=cflags
)

setup(
    name='accelerator-toolbox',
    version='0.1.0',
    description='Accelerator Toolbox',
    long_description=long_description,
    author='The AT collaboration',
    author_email='atcollab-general@lists.sourceforge.net',
    url='https://github.com/atcollab/at',
    # Numpy 1.14.3 is the oldest version that builds with Python 3.7.
    install_requires=['numpy>=1.14.3', 'scipy>=0.16'],
    packages=find_packages(),
    ext_modules=[at, diffmatrix] + [integrator_ext(pm) for pm in pass_methods],
    zip_safe=False,
    python_requires='>=2.7.4',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



"""
Functions relating to fast_ring
"""
import numpy
from math import sqrt, atan2, pi
from at.lattice import uint32_refpts, get_refpts
from at.lattice import Element, RFCavity, Marker, Drift
from at.physics import find_orbit6, find_orbit4
from at.physics import gen_m66_elem, gen_detuning_elem, gen_quantdiff_elem

__all__ = ['fast_ring']


def rearrange(ring, split_inds=[]):
    iends = numpy.concatenate((split_inds, [len(ring)+1]))
    ibegs = numpy.concatenate(([0], iends[:-1]))
    iends = iends[iends != 0]
    ibegs = ibegs[ibegs != len(ring) + 1]
    all_rings = [ring[int(ibeg):int(iend)]
                 for ibeg, iend in zip(ibegs, iends)]

    for ring_slice in all_rings:
        # replace cavity with length > 0 with drift
        # set cavity length to 0 and move to start
        icav = get_refpts(ring_slice, RFCavity)
        for i in numpy.arange(len(icav)):
            cav_elem = ring_slice.pop(int(icav[i]))
            if cav_elem.Length != 0:
                cavdrift = Drift('CavDrift', cav_elem.Length)
                ring_slice.insert(icav[i], cavdrift)
                icav = icav + 1
                cav_elem.Length = 0.0
            ring_slice.insert(0, cav_elem)

        # merge all cavities with the same frequency
        icav = get_refpts(ring_slice, RFCavity)
        all_freq = numpy.array([ring_slice[ic].Frequency for ic in icav])
        all_volt = numpy.array([ring_slice[ic].Voltage for ic in icav])
        uni_freq = numpy.unique(all_freq)

        for ii in numpy.arange(len(uni_freq)):
            fr = uni_freq[ii]
            cavmsk = all_freq == fr
            vol = numpy.sum(all_volt[cavmsk])
            ring_slice[ii].Frequency = fr
            ring_slice[ii].Voltage = vol

        for pp in numpy.arange(len(icav)-len(uni_freq)):
            ring_slice.pop(len(uni_freq))

        ring_slice.insert(len(uni_freq), Marker('xbeg'))
        ring_slice.append(Marker('xend'))
    return all_rings


def merge_rings(all_rings):
    ringnorad = all_rings[0]
    if len(all_rings) > 1:
        for ringi in all_rings[1:]:
            ringnorad += ringi
    ringrad = ringnorad.deepcopy()
    if ringnorad.radiation:
        ringnorad.radiation_off(quadrupole_pass='auto')
    else:
        ringrad.radiation_on(quadrupole_pass='auto')
    return ringnorad, ringrad


def fast_ring(ring, split_inds=[]):
    all_rings = rearrange(ring, split_inds=split_inds)
    ringnorad, ringrad = merge_rings(all_rings)

    ibegs = get_refpts(ringnorad, 'xbeg')
    iends = get_refpts(ringnorad, 'xend')
    markers = numpy.sort(numpy.concatenate((ibegs, iends)))

    _, orbit4 = ringnorad.find_sync_orbit(dct=0.0, refpts=markers)
    _, orbit6 = ringrad.find_orbit6(refpts=markers)

    detuning_elem = gen_detuning_elem(ringnorad, orbit4[-1])
    detuning_elem_rad = detuning_elem.deepcopy()
    detuning_elem_rad.T1 = -orbit6[-1]
    detuning_elem_rad.T2 = orbit6[-1]

    for counter, ring_slice in enumerate(all_rings):

        ibeg = get_refpts(ring_slice, 'xbeg')[0]
        iend = get_refpts(ring_slice, 'xend')[0]
        cavs = ring_slice[:ibeg]
        ring_slice = ring_slice[ibeg:iend]
        ring_slice_rad = ring_slice.deepcopy()

        if ring_slice.radiation:
            ring_slice.radiation_off(quadrupole_pass='auto')
        else:
            ring_slice_rad.radiation_on(quadrupole_pass='auto')

        lin_elem = gen_m66_elem(ring_slice, orbit4[2*counter],
                                orbit4[2*counter+1])
        lin_elem.FamName = lin_elem.FamName + '_' + str(counter)

        qd_elem = gen_quantdiff_elem(ring_slice_rad, orbit=orbit6[2*counter])
        qd_elem.FamName = qd_elem.FamName + '_' + str(counter)
        lin_elem_rad = gen_m66_elem(ring_slice_rad, orbit6[2*counter],
                                    orbit6[2*counter+1])
        lin_elem_rad.FamName = lin_elem_rad.FamName + '_' + str(counter)

        [ringnorad.append(cav) for cav in cavs]
        ringnorad.append(lin_elem)
        [ringrad.append(cav) for cav in cavs]
        ringrad.append(lin_elem_rad)
        ringrad.append(qd_elem)

    ringnorad.append(detuning_elem)
    ringrad.append(detuning_elem_rad)
    del ringnorad[:markers[-1]+1]
    del ringrad[:markers[-1]+1]

    return ringnorad, ringrad



from itertools import chain
import numpy as np
from scipy.optimize import least_squares
from itertools import repeat
from at.lattice import refpts_iterator, bool_refpts, uint32_refpts
from at.physics import linopt, ohmi_envelope, find_orbit4


class Variable(object):
    """A Variable is a scalar value acting on a lattice through the
    user-defined functions setfun and getfun
    """
    def __init__(self, setfun, getfun, name='', bounds=(-np.inf, np.inf),
                 *args, **kwargs):
        self.setfun = setfun
        self.getfun = getfun
        self.name = name
        self.bounds = bounds
        self.args = args
        self.kwargs = kwargs
        super(Variable, self).__init__()

    def set(self, ring, value):
        self.setfun(ring, value, *self.args, **self.kwargs)

    def get(self, ring):
        return self.getfun(ring, *self.args, **self.kwargs)

    @staticmethod
    def header():
        return '\n{:>12s}{:>13s}{:>16s}{:>16s}\n'.format(
            'Name', 'Initial', 'Final ', 'Variation')

    def status(self, ring, vini=np.NaN):
        vnow = self.get(ring)
        return '{:>12s}{: 16e}{: 16e}{: 16e}'.format(
            self.name, vini, vnow, (vnow - vini) / vini)


class ElementVariable(Variable):
    """An ElementVariable is:
    - a scalar attribute or
    - an element of an array attribute
    of one or several elements of a lattice"""

    def __init__(self, refpts, attname, index=None, **kwargs):
        setf, getf = self._access(index)

        def setfun(ring, value):
            for elem in refpts_iterator(ring, refpts):
                setf(elem, attname, value)

        def getfun(ring):
            values = np.array([getf(elem, attname) for elem in
                               refpts_iterator(ring, refpts)])
            return np.average(values)

        super(ElementVariable, self).__init__(setfun, getfun, **kwargs)
        self.refpts = refpts

    @staticmethod
    def _access(index):
        """Access to element attributes"""
        if index is None:
            setf = setattr
            getf = getattr
        else:
            def setf(elem, attrname, value):
                getattr(elem, attrname)[index] = value

            def getf(elem, attrname):
                return getattr(elem, attrname)[index]
        return setf, getf


class Constraints(object):
    """Container for generic constraints:
      - a constraint is defined by a user-defined evaluation function.
      - constraints are added to the container with the Constraints.add method

      Example:
          # define an evaluation function for the ring circumference:
          def circ_fun(ring):
              return ring.get_s_pos(len(ring) + 1)

          # define an evaluation function for the momentum compaction factor:
          def mcf_fun(ring):
              return ring.get_mcf()

          # Construct the container:
          cnstrs = Constraints()

          # Add the two constraints:
          cnstrs.add(circ_fun, 850.0)
          cnstrs.add(mcf_fun, 1.0e-4, weight=0.1)
    """
    def __init__(self, *args, **kwargs):
        """Constraints(*args, **kwargs)
        build a generic constraints container.

        The positional and keyword parameters are provided to all the evaluation
        functions.
        """
        self.name = []
        self.fun = []
        self.target = []
        self.weight = []
        self.lbound = []
        self.ubound = []
        self.rad = kwargs.pop('rad', False)
        self.args = args
        self.kwargs = kwargs

    def add(self, fun, target, name=None, weight=1.0, bounds=(0.0, 0.0)):
        """Add a target to the Constraints container

        PARAMETERS
            fun           evaluation function. Called as:
                          value = fun(ring, *args, **kwargs)
                            value is the constrained parameter value
                            value may be a scalar or an array.
                            the positional and keyword parameters come from
                            the Constraints initialisation
            target        desired value.

        KEYWORDS
            name=None     name of the constraint. If None, name is generated
                          from the name of the evaluation function
            weight=1.0    weight factor: the residual is (value-target)/weight
            bounds=(0,0)  lower and upper bounds. The parameter is constrained
                          in the interval [target-low_bound target+up_bound]

        The "target", "weight" and "bounds" input must be broadcastable to the
        shape of "value".
        """
        if name is None:                # Generate the constraint name
            name = fun.__name__
        self.name.append(name)
        self.fun.append(fun)
        target, lbound, ubound = np.atleast_1d(target, *bounds)
        self.target.append(target)
        self.weight.append(weight)
        self.lbound.append(lbound)
        self.ubound.append(ubound)

    def values(self, ring):
        """Return the list of actual parameter values"""
        return [fun(ring, *self.args, **self.kwargs) for fun in self.fun]

    def evaluate(self, ring):
        """Return a flattened array of weighted residuals"""

        def res(value, target, weight, lbound, ubound):
            diff = value - target   # broadcast here => at least 1 dimension
            lb = diff - lbound
            ub = diff - ubound
            lb[lb >= 0] = 0         # needs at least 1 dimension
            ub[ub <= 0] = 0
            return np.maximum(abs(lb), abs(ub)) / weight

        return np.concatenate([res(v, t, w, lb, ib) for v, t, w, lb, ib
                               in zip(self.values(ring), self.target,
                                      self.weight, self.lbound,
                                      self.ubound)], axis=None)

    @staticmethod
    def header():
        """Header for the display of constraint values and residuals"""
        return '\n{:>12s}{:>13s}{:>16s}{:>16s}{:>16s}\n'.format(
            'Name', 'Initial', 'Final ', 'Target', 'Residual')

    def status(self, ring, initial=None):
        """Return a string giving the actual state of constraints"""
        if initial is None:
            initial = repeat(np.NaN)
        strs = []
        for name, ini, now, target in zip(self.name, initial,
                                          self.values(ring), self.target):
            now, target = np.broadcast_arrays(now, target)
            for vi, vn, vt in zip(np.broadcast_to(ini, now.shape).flat,
                                  now.flat,
                                  np.broadcast_to(target, now.shape).flat):
                strs.append('{:>12s}{: 16e}{: 16e}{: 16e}{: 16e}'.format(
                    name, vi, vn, vt, vn - vt))
        return '\n'.join(strs)


class ElementConstraints(Constraints):
    """Base class for position-related constraints: handle the refpoints
    of each target"""
    def __init__(self, ring, *args, **kwargs):
        self.nelems = len(ring)
        self.refs = []
        self.refpts = bool_refpts([], self.nelems)
        super(ElementConstraints, self).__init__(*args, **kwargs)

    @staticmethod
    def _arrayaccess(index):
        """Access to array elements"""
        if index is None:
            def getv(x):
                return x
        else:
            def getv(x):
                return x[index]
        return getv

    @staticmethod
    def _recordaccess(index):
        """Access to optics parameters"""
        if index is None:
            getf = getattr
        else:
            if isinstance(index, tuple):
                idx = (Ellipsis,)+index
            else:
                idx = (Ellipsis, index)

            def getf(lindata, attrname):
                return getattr(lindata, attrname)[idx]
        return getf

    def add(self, fun, target, refpts=None, **kwargs):
        ref = bool_refpts(refpts, self.nelems)
        # Store the new refpoint
        self.refs.append(ref)
        # Update the union of all refpoints
        self.refpts = np.stack((self.refpts, ref), axis=0).any(axis=0)
        super(ElementConstraints, self).add(fun, target, **kwargs)

    def values(self, ring):
        # Single optics computation
        vloc, vglob = self.compute(ring, *self.args, **self.kwargs)
        # Evaluate all constraints
        return [fun(loc, *vglob) for fun, loc in zip(self.fun, vloc)]

    def compute(self, ring, *args, **kwargs):
        """Dummy computation. Compute must return:
        - an iterator over local data for each target
        - a tuple of global data"""
        return repeat(None), ()


class LinoptConstraints(ElementConstraints):
    """Container for linear optics constraints:
      - a constraint can be set on any result of at.linopt
      - constraints are added to the container with the LinoptConstraints.add
        method.

      at.linopt is called once before the evaluation of all constraints

      Example:
          cnstrs = LinoptConstraints(ring, dp=0.01, coupled=False)

          # Add a beta H (beta[0]) constraint at location ref_inj
          cnstrs.add('beta_x_inj', 'beta', 18.0, refpts=ref_inj, index=0)

          # Add a tune constraint
          cnstrs.add('tune', 0.44, index=0, weight=0.01)

          # Add a chromaticity constraint (both planes)
          cnstrs.add('chrom', [0.0 0.0])

          # define a constraint of phase advances between 2 points
          def mu_diff(lindata, tune, chrom):
              delta_mu = (lindata[1].mu - lindata[0].mu)/(2*np.pi)
              return delta_mu % 1.0

          # Add a H phase advance constraint, giving the desired locations
          cnstrs.add(mu_diff, 0.5, refpts=[sf0 sf1], index=0)
    """
    def __init__(self, ring, **kwargs):
        """Build a LinoptConstraints container

        KEYWORDS
        dp=0.0          momentum deviation.
        coupled=True    if False, simplify the calculations by assuming
                        no H/V coupling
        twiss_in=None   Initial twiss parameters for transfer line optics.
                        "lindata" stucture, where only the beta and alpha are
                        required and used.
        orbit=None      Initial trajectory for transfer line
                        ((6,) array)
        """
        self.get_chrom = False
        super(LinoptConstraints, self).__init__(ring, **kwargs)

    def add(self, param, target, refpts=None, index=None, name=None, **kwargs):
        """Add a target to the LinoptConstraints container

        PARAMETERS
            param         2 possibilities:
                          - parameter name: see at.linopt for the name of
                            available parameters. In addition to local optical
                            parameters, 'tunes' and 'chroms' are allowed.
                          - user-supplied parameter evaluation function:
                                value = param(lindata, tune, chrom)
                            lindata contains the optics parameters at all the
                              specified refpoints
                            value is the constrained parameter value
                              (scalar or array).
            target        desired value.

        KEYWORDS
            refpts=None   location of the constraint. Several locations may be
                          given to apply the same constraint at several points.
            index=None    index in the parameter array. If None, the full array
                          is used.
            name=None     name of the constraint. If None, name is generated
                          from param and index.
            weight=1.0    weight factor: the residual is (value-target)/weight.
            bounds=(0,0)  lower and upper bounds. The parameter is constrained
                          in the interval [target-low_bound target+up_bound]

        The target, weight and bounds values must be broadcastable to the shape
        of value.
        """
        getf = self._recordaccess(index)
        getv = self._arrayaccess(index)

        if name is None:                # Generate the constraint name
            name = param.__name__ if callable(param) else param
            if index is not None:
                name = '{}_{}'.format(name, index)

        if callable(param):
            def fun(refdata, tune, chrom):
                return getv(param(refdata, tune, chrom))
            # self.refpts[:] = True     # necessary not to miss 2*pi jumps
            self.get_chrom = True       # fun may use dispersion or chroma
        elif param == 'tunes':
            # noinspection PyUnusedLocal
            def fun(refdata, tune, chrom):
                return getv(tune)
            refpts = []
        elif param == 'chroms':
            # noinspection PyUnusedLocal
            def fun(refdata, tune, chrom):
                return getv(chrom)
            refpts = []
            self.get_chrom = True       # slower but necessary
        else:
            # noinspection PyUnusedLocal
            def fun(refdata, tune, chrom):
                return getf(refdata, param)
            if param == 'dispersion':
                self.get_chrom = True   # slower but necessary
            # elif param == 'mu':
            #     self.refpts[:] = True # necessary not to miss 2*pi jumps

        super(LinoptConstraints, self).add(fun, target, refpts, name=name,
                                           **kwargs)

    def compute(self, ring, *args, **kwargs):
        """Optics computation before evaluation of all constraints"""
        ld0, tune, chrom, ld = linopt(ring, refpts=self.refpts,
                                      get_chrom=self.get_chrom, **kwargs)
        return (ld[ref[self.refpts]] for ref in self.refs), (tune, chrom)


class Orbit4Constraints(ElementConstraints):
    """Container for orbit constraints:
    The closed orbit can be handled with LinoptConstraints, but for problems
    which do not involve parameters other than orbit, like steering or
    orbit bumps, Orbit4Constraints is much faster.

      at.find_orbit4 is called once before the evaluation of all constraints

    Example:
        cnstrs = Orbit4Constraints(ring, dp=0.01)

        # Add a bump (x=-0.004, x'=0) constraint at location ref_inj
        cnstrs.add([-0.004, 0.0], refpts=ref_inj, index=slice(2))
    """
    def __init__(self, ring, *args, **kwargs):
        """Build a Orbit4Constraints container

        KEYWORDS
        dp=0.0              momentum deviation.
        orbit=None          Initial trajectory for transfer line: (6,) array
        """
        super(Orbit4Constraints, self).__init__(ring, *args, **kwargs)

    def add(self, target, refpts=None, index=None, name=None, **kwargs):
        """Add a target to the Orbit4Constraints container

        PARAMETERS
            target        desired value.

        KEYWORDS
            refpts=None   location of the constraint. Several locations may be
                          given to apply the same constraint at several points.
            index=None    index in the orbit vector. If None, the full orbit
                          is used. Example:
                            index=0         # x
                            index=2         # z
                            index=slice(4)  # x, x', z, z'
            name='orbit4' name of the constraint.
            weight=1.0    weight factor: the residual is (value-target)/weight.
            bounds=(0,0)  lower and upper bounds. The parameter is constrained
                          in the interval [target-low_bound target+up_bound]

        The target, weight and bounds values must be broadcastable to the shape
        of value.
        """

        if name is None:                # Generate the constraint name
            name = 'orbit4' if index is None else 'orbit4_{}'.format(index)

        fun = self._arrayaccess(index)
        super(Orbit4Constraints, self).add(fun, target, refpts, name=name,
                                           **kwargs)

    def compute(self, ring, *args, **kwargs):
        """Orbit computation before evaluation of all constraints"""
        orbit0, orbit = find_orbit4(ring, refpts=self.refpts, **kwargs)
        return (orbit[ref[self.refpts]].T for ref in self.refs), ()


class EnvelopeConstraints(ElementConstraints):
    """Container for envelope constraints:
      - a constraint can be set on any result of at.ohmi_envelope
      - constraints are added to the container with the EnvelopeConstraints.add
        method.

      at.ohmi_envelope is called once before the evaluation of all constraints

      Example:
          cnstrs = EnvelopeConstraints(ring)
    """
    def __init__(self, ring):
        """Build a EnvelopeConstraints container"""
        super(EnvelopeConstraints, self).__init__(ring, rad=True)

    def add(self, param, target, refpts=None, index=None, name=None, **kwargs):
        """Add a target to the EnvelopeConstraints container

        PARAMETERS
            param         2 possibilities:
                          - parameter name: see at.ohmi_envelope for the name of
                            available parameters. In addition to local
                            parameters, 'tunes', 'damping_rates',
                            'mode_matrices' and 'mode_emittance' are allowed.
                          - user-supplied parameter evaluation function:
                                value = prm(emit_data, beam_data)
                            emit_data contains the emittance data at all the
                              specified refpoints
                            value is the constrained parameter value
                              (scalar or array).
            target        desired value.

        KEYWORDS
            refpts=None   location of the constraint. Several locations may be
                          given to apply the same constraint at several points.
            index=None    index in the parameter array. If None, the full array
                          is used.
            name=None     name of the constraint. If None, name is generated
                          from param and index.
            weight=1.0    weight factor: the residual is (value-target)/weight.
            bounds=(0,0)  lower and upper bounds. The parameter is constrained
                          in the interval [target-low_bound target+up_bound]

        The target, weight and bounds values must be broadcastable to the shape
        of value.
        """
        def beam_constraint(prm):
            # noinspection PyUnusedLocal
            def beamfunc(refdata, beam_data):
                return getv(beam_data[prm])
            return beamfunc

        getf = self._recordaccess(index)
        getv = self._arrayaccess(index)

        if name is None:                # Generate the constraint name
            name = param.__name__ if callable(param) else param
            if index is not None:
                name = '{}_{}'.format(name, index)

        if callable(param):
            def fun(refdata, beam_data):
                return getv(param(refdata, beam_data))
        elif param in ['tunes', 'damping_rates', 'mode_matrices',
                       'mode_emittances']:
            fun = beam_constraint(param)
            refpts = []
        else:
            # noinspection PyUnusedLocal
            def fun(refdata, beam_data):
                return getf(refdata, param)

        super(EnvelopeConstraints, self).add(fun, target, refpts, name=name,
                                             **kwargs)

    def compute(self, ring, *args, **kwargs):
        """Optics computation before evaluation of all constraints"""
        em0, beamdata, em = ohmi_envelope(ring, refpts=self.refpts, **kwargs)
        return (em[ref[self.refpts]] for ref in self.refs), (beamdata,)


def match(ring, variables, constraints, verbose=2, max_nfev=1000,
          diff_step=1.0e-10, method=None):
    """Perform matching of constraints by varying variables

    PARAMETERS
        ring                ring lattice or transfer line
        variables           sequence of Variable objects
        constraints         sequance of Constraints objects

    KEYWORDS
        verbose=2           See scipy.optimize.least_squares
        max_nfev=1000           "
        diff_step=1.0e-10       "
    """
    def fun(vals):
        for value, variable in zip(vals, variables):
            variable.set(ring1, value)
            variable.set(ring2, value)

        c1 = [cons.evaluate(ring1) for cons in cst1]
        c2 = [cons.evaluate(ring2) for cons in cst2]
        return np.concatenate(c1 + c2, axis=None)

    cst1 = [cons for cons in constraints if not cons.rad]
    cst2 = [cons for cons in constraints if cons.rad]

    ring1 = ring.copy()         # Make a shallow copy of ring
    for var in variables:       # Make a deep copy of varying elements
        for ref in uint32_refpts(var.refpts, len(ring1)):
            ring1[ref] = ring1[ref].deepcopy()

    ring2 = ring1.radiation_on(copy=True)

    aaa = [(var.get(ring1), var.bounds) for var in variables]
    vini, bounds = zip(*aaa)
    bounds = np.array(bounds).T

    cini1 = [cst.values(ring1) for cst in cst1]
    cini2 = [cst.values(ring2) for cst in cst2]
    ntargets = sum(np.size(a) for a in chain.from_iterable(cini1 + cini2))

    if method is None:
        if np.all(abs(bounds) == np.inf) and ntargets >= len(variables):
            method = 'lm'
        else:
            method = 'trf'
    if verbose >= 1:
        print('\n{} constraints, {} variables, using method {}\n'.
              format(ntargets, len(variables), method))

    least_squares(fun, vini, bounds=bounds, verbose=verbose, max_nfev=max_nfev,
                  method=method, diff_step=diff_step)

    if verbose >= 1:
        print(Constraints.header())
        for cst, ini in zip(cst1, cini1):
            print(cst.status(ring1, initial=ini))
        for cst, ini in zip(cst2, cini2):
            print(cst.status(ring2, initial=ini))

        print(Variable.header())
        for var, vini in zip(variables, vini):
            print(var.status(ring1, vini=vini))

    return ring1
